# -*- coding: utf-8 -*-

from __future__ import division


def bateman_parent(lmbd, t, one=1, zero=0, exp=None):
    """ Calculate daughter concentrations (number densities) from single parent

    Assumes an initial parent concentraion of one (and zero for all daughters)

    Parameters
    ----------
    lmbd: array_like
        decay constants (one per species)
    t: float
        time
    one: object
        object corresponding to one, default(int(1)),
        could e.g. be :py:func:`sympy.S('1')`
    zero: object
        object corresponding to zero, default(int(0))
        could e.g. be :py:func:`sympy.S('0')`
    exp: callback
        Callback calculating the exponential of an argument
        default: numpy.exp or math.exp, could e.g. be :py:func:`sympy.exp`

    """
    n = len(lmbd)
    N = [None]*n
    lmbd_prod = one
    if exp is None:
        try:
            from numpy import exp
        except ImportError:
            from math import exp
    for i in range(n):
        if i > 0:
            lmbd_prod *= lmbd[i-1]
        sum_k = zero
        for k in range(i+1):
            prod_l = one
            for l in range(i+1):
                if l == k:
                    continue
                prod_l *= lmbd[l] - lmbd[k]
            sum_k += exp(-lmbd[k]*t)/prod_l
        N[i] = lmbd_prod*sum_k
    return N


def bateman_full(y0s, lmbd, t, one=1, zero=0, exp=None):
    """ Calculates a linear combination of single-parent chains

    Generalized helper function for when y0 != [1, 0, 0, ... ]

    Parameters
    ----------
    y0s: array_like
        Initial concentrations
    t: float
        time
    one: object
        object corresponding to one, default(int(1)),
        could e.g. be :py:func:`sympy.S('1')`
    zero: object
        object corresponding to zero, default(int(0))
        could e.g. be :py:func:`sympy.S('0')`
    exp: callback
        Callback calculating the exponential of an argument
        default: math.exp, could e.g. be :py:func:`sympy.exp`

    """
    n = len(lmbd)
    if len(y0s) != n:
        raise ValueError("Please pass equal number of decay"
                         " constants as initial concentrations"
                         " (you may want to pad lmbd with zeroes)")
    N = [zero]*n
    for i, y0 in enumerate(y0s):
        if y0 == zero:
            continue
        Ni = bateman_parent(lmbd[i:], t, one, zero, exp)
        for j, yj in enumerate(Ni, i):
            N[j] += y0*yj
    return N



# -*- coding: utf-8 -*-
""" batemaneq provides a Python package for evaluating the Bateman equation """
from __future__ import absolute_import

from ._release import __version__
from .bateman import bateman_parent, bateman_full
from ._bateman_double import bateman_parent as bateman_parent_arr
from ._bateman_double import bateman_full as bateman_full_arr



# -*- coding: utf-8 -*-
from __future__ import division
import numpy as np
from batemaneq import bateman_full, bateman_full_arr, bateman_parent
import pytest

decay_analytic = {
    0: lambda y0, k, t: (
        y0[0] * np.exp(-k[0]*t)),
    1: lambda y0, k, t: (
        y0[1] * np.exp(-k[1] * t) + y0[0] * k[0] / (k[1] - k[0]) *
        (np.exp(-k[0]*t) - np.exp(-k[1]*t))),
    2: lambda y0, k, t: (
        y0[2] * np.exp(-k[2] * t) + y0[1] * k[1] / (k[2] - k[1]) *
        (np.exp(-k[1]*t) - np.exp(-k[2]*t)) +
        k[1] * k[0] * y0[0] / (k[1] - k[0]) *
        (1 / (k[2] - k[0]) * (np.exp(-k[0]*t) - np.exp(-k[2]*t)) -
         1 / (k[2] - k[1]) * (np.exp(-k[1]*t) - np.exp(-k[2]*t))))
}


def decay_get_Cref(k, y0, tout):
    coeffs = k + [0]*(3-len(k))
    return np.column_stack([
        decay_analytic[i](y0, coeffs, tout) for i in range(
            min(3, len(k)+1))])


def test_bateman_full():
    k0, k1, k2 = 2.0, 3.0, 4.0
    k = [k0, k1, k2]
    y0 = [0.7, 0.3, 0.5]
    t = np.linspace(0, 10, 17)
    yout = bateman_full(y0, k, t, exp=np.exp)
    yref = decay_get_Cref(k, y0, t)
    assert np.allclose(np.asarray(yout).T, yref)


def test_bateman_full_arr():
    k0, k1, k2 = 2.0, 3.0, 4.0
    k = [k0, k1, k2]
    y0 = [0.7, 0.3, 0.5]
    t = np.linspace(0, 10, 17)
    yout = bateman_full_arr(np.asarray(y0), np.asarray(k), t)
    yref = decay_get_Cref(k, y0, t)
    assert np.allclose(yout, yref)


def _yi1(i, p, a, binom):
    return binom(p+i-1, p) * a**(-1-p) * ((a-1)/a)**(i-1)


@pytest.mark.parametrize('p', (0, 1, 2, 3, 4, 5))
def test_basteman_parent(p):
    from scipy.special import binom
    from math import log
    N = 32
    a = 27
    lmbd = [(i+p+1)*log(a) for i in range(N)]
    bp = bateman_parent(lmbd, 1)
    for i, v in enumerate(bp, 1):
        assert abs(v - _yi1(i, p, a, binom)) < 1e-14



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Tested with boost GCC 4.8.3

import os
import shutil
import sys
from distutils.core import setup
from distutils.extension import Extension
import numpy as np


pkg_name = 'batemaneq'

# Cythonize .pyx file if it exists (not in source distribution)
ext_modules = []

if len(sys.argv) > 1 and '--help' not in sys.argv[1:] and sys.argv[1] not in (
            '--help-commands', 'egg_info', 'clean', '--version'):
    USE_CYTHON = os.path.exists('batemaneq/_bateman_double.pyx')
    ext = '.pyx' if USE_CYTHON else '.cpp'
    sources = ['src/bateman_double.cpp', 'batemaneq/_bateman_double'+ext]
    ext_modules = [
        Extension('batemaneq._bateman_double', sources, language='c++',
                  extra_compile_args=['-std=c++11'],
                  include_dirs=['./include'])
    ]
    if USE_CYTHON:
        from Cython.Build import cythonize
        ext_modules = cythonize(ext_modules, include_path=['./include'],
                                gdb_debug=True)

BATEMANEQ_RELEASE_VERSION = os.environ.get('BATEMANEQ_RELEASE_VERSION', '')

# http://conda.pydata.org/docs/build.html#environment-variables-set-during-the-build-process
CONDA_BUILD = os.environ.get('CONDA_BUILD', '0') == '1'
if CONDA_BUILD:
    try:
        BATEMANEQ_RELEASE_VERSION = 'v' + open(
            '__conda_version__.txt', 'rt').readline().rstrip()
    except IOError:
        pass

release_py_path = os.path.join(pkg_name, '_release.py')

if len(BATEMANEQ_RELEASE_VERSION) > 1 and BATEMANEQ_RELEASE_VERSION[0] == 'v':
    TAGGED_RELEASE = True
    __version__ = BATEMANEQ_RELEASE_VERSION[1:]
else:
    TAGGED_RELEASE = False
    # read __version__ attribute from release.py:
    exec(open(release_py_path).read())

classifiers = [
    "Development Status :: 4 - Beta",
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Mathematics',
]

tests = [
    'batemaneq.tests',
]

long_description = open('README.rst').read()
with open(os.path.join(pkg_name, '__init__.py')) as f:
    short_description = f.read().split('"""')[1]

setup_kwargs = dict(
    name=pkg_name,
    version=__version__,
    description=short_description,
    classifiers=classifiers,
    author='Björn Dahlgren',
    author_email='bjodah@DELETEMEgmail.com',
    license='BSD',
    url='https://github.com/bjodah/' + pkg_name,
    packages=[pkg_name] + tests,
    ext_modules=ext_modules,
    include_dirs=[np.get_include(), './include']
)

if __name__ == '__main__':
    try:
        if TAGGED_RELEASE:
            # Same commit should generate different sdist
            # depending on tagged version (set BATEMANEQ_RELEASE_VERSION)
            # this will ensure source distributions contain the correct version
            shutil.move(release_py_path, release_py_path+'__temp__')
            open(release_py_path, 'wt').write(
                "__version__ = '{}'\n".format(__version__))
        setup(**setup_kwargs)
    finally:
        if TAGGED_RELEASE:
            shutil.move(release_py_path+'__temp__', release_py_path)
